/*******************************************************************************
  MPLAB Harmony Application Source File

  Company:
    Microchip Technology Inc.

  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It
    implements the logic of the application's state machine and it may call
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#define ECB 1
#define AES128 1

#include "tiny-AES-c/aes.h"
#include "app.h"
#include "cryptoauthlib.h"
#include "host/atca_host.h"
#include "../../slot_5_secret_key.h"

// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

#define SYMMETRIC_KEY_SLOT  (0x05)

extern ATCAIfaceCfg atecc608_0_init_data;
extern uint8_t slot_5_secret_key[];

uint8_t g_salt[16];
uint8_t g_mcu_ephemeral_key[16];
uint8_t g_remote_host_ephemeral_key[16];
uint8_t g_plaintext[16];
uint8_t g_ciphertext[16];

// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    This structure should be initialized by the APP_Initialize function.

    Application strings and buffers are be defined outside this structure.
*/

APP_DATA appData;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

/* TODO:  Add any necessary callback functions.
*/

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************

/* Generate random numbers using secure element */
void RNG(uint8_t *dest, size_t size)
{
    ATCA_STATUS status = ATCA_GEN_FAIL;
    uint8_t random[RANDOM_NUM_SIZE];

    if (ATCA_SUCCESS == (status = atcab_random(random)))
    {
        memcpy(dest, random, size);
    }
}

/* Initialize secure element and generate random numbers for plaintext and KDF salt */
ATCA_STATUS aes_app_init(void)
{
    ATCA_STATUS status;
    
    char displaystr[400];
    size_t displaylen;
    
    printf("\n\r------- AES Message Encryption Usecase ------\n\n\r");
    printf("STEP 1 - Device Initialization\n\r");
    status = atcab_init(&atecc608_0_init_data);
    if (status != ATCA_SUCCESS){
        printf("ECC608 Initialization failed with status {%02X}.\r\n", status);
        return status;
    }
    printf("ECC608 Initialization - successful\n\r");
    
    // Generating 16 byte random number to use as salt for KDF operation
    printf("Generating random number to be used as Salt for KDF\n\r");
    RNG(g_salt, sizeof(g_salt)/sizeof(g_salt[0]));
    displaylen = sizeof(displaystr);
    status = atcab_bin2hex(g_salt, sizeof(g_salt)/sizeof(g_salt[0]), displaystr, &displaylen);
    printf("\nRandom Salt:\r\n%s\r\n", displaystr);
    
    // Generating 16 byte random number to use as plain text
    printf("Generating random number to be used as plaintext\n\r");
    RNG(g_plaintext, sizeof(g_plaintext)/sizeof(g_plaintext[0]));
    displaylen = sizeof(displaystr);
    status = atcab_bin2hex(g_plaintext, sizeof(g_plaintext)/sizeof(g_plaintext[0]), displaystr, &displaylen);
    printf("\nGenerated plaintext:\r\n%s\r\n", displaystr);
    
    return status;
}

/* Derive ephemeral key using secure element to use for encryption/decryption */
ATCA_STATUS derive_key_mcu_ecc608(void)
{
    ATCA_STATUS status;

    uint8_t kdf_out[32];
    char displaystr[400];
    size_t displaylen;    
    
    printf("\n\r---------------------------------------------\n\r");
    printf("\n\rSTEP 2 - Key derivation on ECC608\n\r");
        
    // Begin KDF operation
    // Running KDF in AES mode, slot 5 as source key and output is sent on the I2C bus
    // Addition data is 0 as the first key location in Slot 5 is used
    // Out_nonce is not required as output is read as plain text
    status = atcab_kdf(
        KDF_MODE_ALG_AES | KDF_MODE_SOURCE_SLOT | KDF_MODE_TARGET_OUTPUT,
        0x0005,
        0,
        g_salt,
        kdf_out,
        NULL);
    if (status == ATCA_SUCCESS)
    {
        memcpy(g_mcu_ephemeral_key, kdf_out, 16);
        displaylen = sizeof(displaystr);
        atcab_bin2hex(kdf_out, 16, displaystr, &displaylen);
        printf("\nGenerated Ephemeral Key (Derived key) on MCU:\r\n%s\r\n", displaystr);
        printf("MCU can now use this key to do AES Encrypt/Decrypt\r\n");
    }
    else
    {
        printf("\nDevice KDF failed with status {%02X}.\r\n", status);
    }    
    
    return status;
}

/* Encrypt plaintext using software AES */
ATCA_STATUS encrypt_msg_on_mcu(void)
{
    ATCA_STATUS status;
    
    uint8_t ciphertext[16];
    char displaystr[400];
    size_t displaylen;    
    
    printf("\n\r---------------------------------------------\n\r");
    printf("\n\rSTEP 3 - Generate random msg and run encrypt on MCU\n\r");

    printf("\nEncrypting plaintext using tiny-AES-c software AES library\r\n");
    // Running AES encryption using tiny-AES-c software AES library
    memcpy(ciphertext, g_plaintext, 16);
    struct AES_ctx ctx;
    AES_init_ctx(&ctx, g_mcu_ephemeral_key);
    AES_ECB_encrypt(&ctx, ciphertext);
    memcpy(g_ciphertext, ciphertext, 16);
    displaylen = sizeof(displaystr);
    status = atcab_bin2hex(ciphertext, sizeof(ciphertext)/sizeof(ciphertext[0]), displaystr, &displaylen);
    printf("\nEncrypted message:\r\n%s\r\n", displaystr);    
    return status;
}

/* Derive ephemeral key on host for encryption/decryption */
ATCA_STATUS derive_key_on_remote_host(void)
{
    ATCA_STATUS status = ATCA_SUCCESS;

    uint8_t host_derive_key_buf[16];
    char displaystr[400];
    size_t displaylen;        
    
    memcpy(host_derive_key_buf, g_salt, 16);
    printf("\n\r---------------------------------------------\n\r");
    printf("\n\rSTEP 4 - Key derivation on Remote Host\n\r");
    
    printf("Use salt and master key to generate derived key on Remote Host\n\r");
    struct AES_ctx ctx;
    AES_init_ctx(&ctx, slot_5_secret_key);
    AES_ECB_encrypt(&ctx, host_derive_key_buf);
    displaylen = sizeof(displaystr);
    status = atcab_bin2hex(host_derive_key_buf, sizeof(host_derive_key_buf)/sizeof(host_derive_key_buf[0]), displaystr, &displaylen);
    printf("\nGenerated Ephemeral key on Remote host:\r\n%s\r\n", displaystr);
    printf("Remote Host can now use this key to do AES Encrypt/Decrypt\r\n");
    memcpy(g_remote_host_ephemeral_key, host_derive_key_buf, 16);
    return status;    
}

/* Decrypt encrypted message and compare with original plaintext */
ATCA_STATUS decrypt_and_compare_msg_remote_host(void)
{
    ATCA_STATUS status = ATCA_SUCCESS;
    
    uint8_t decrypt_buf[16];
    char displaystr[400];
    size_t displaylen;
    
    printf("\n\r---------------------------------------------\n\r");
    printf("\n\rSTEP 5 - Decrypting the encrypted message on Remote Host\n\r");
    
    struct AES_ctx ctx;
    AES_init_ctx(&ctx, g_remote_host_ephemeral_key);
    // AES_ECB_decrypt will operate on input buffer, copying cipher text to decrypt_buf
    memcpy(decrypt_buf, g_ciphertext, 16);
    AES_ECB_decrypt(&ctx, decrypt_buf);
    displaylen = sizeof(displaystr);
    status = atcab_bin2hex(decrypt_buf, sizeof(decrypt_buf)/sizeof(decrypt_buf[0]), displaystr, &displaylen);
    printf("\nDecrypted message:\r\n%s\r\n", displaystr);
    
    printf("\n\rComparing decrypted message against the original message");
    if (memcmp(decrypt_buf, g_plaintext, 16) == 0)
    {
        status = ATCA_SUCCESS;
        printf("\n\rMatch - Message decryption SUCCESSFUL\r\n");
    }
    else
    {
        status = ATCA_FUNC_FAIL;
        printf("\n\rThe plaintext does not match - Message decryption FAILED\r\n");
    }
    
    printf("\n\r---------------------------------------------\n\r\n\r");
    return status;    
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize ( void )
{
    /* Place the App state machine in its initial state. */
    appData.state = APP_STATE_INIT;



    /* TODO: Initialize your application's state machine and other
     * parameters.
     */
}


/******************************************************************************
  Function:
    void APP_Tasks ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks ( void )
{
    ATCA_STATUS status;

    /* Check the application's current state. */
    switch ( appData.state )
    {
        /* Application's initial state. */
        case APP_STATE_INIT:
            status = aes_app_init();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_DERIVE_KEY_MCU : APP_STATE_FINISH;
            break;

        case APP_STATE_DERIVE_KEY_MCU:
            status = derive_key_mcu_ecc608();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_ENCRYPT_MSG : APP_STATE_FINISH;
            break;

        case APP_STATE_ENCRYPT_MSG:
            status = encrypt_msg_on_mcu();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_DERIVE_KEY_HOST : APP_STATE_FINISH;
            break;

        case APP_STATE_DERIVE_KEY_HOST:
            status = derive_key_on_remote_host();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_DECRYPT_MSG : APP_STATE_FINISH;
            break;

        case APP_STATE_DECRYPT_MSG:
            (void)decrypt_and_compare_msg_remote_host();
            appData.state = APP_STATE_FINISH;
            break;

        /* The default state should never be executed. */
        default:
        {
            /* TODO: Handle error in application's state machine. */
            break;
        }
    }
}


/*******************************************************************************
 End of File
 */
